'''
..
  $Id$
  
  HTTP python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 19.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

from clazzes.http.response import DEFAULT_BLOCK_SIZE, send_error

import http.client as httplib
from urllib.parse import quote as url_quote, ParseResult

import socket
import types

# Remove these headers from response (specify lower case header
# names):
FILTERED_HEADERS = (
    'transfer-encoding',
)

def proxy_exact_request(environ:map, start_response:types.FunctionType, gateway_url:ParseResult, block_size=DEFAULT_BLOCK_SIZE):
    """
    HTTP proxying WSGI application that proxies the exact request
    given in the environment.  All controls are passed through the
    environment.

    This connects to the server given in the hostname and port member of the
    ``gateway_url`` parameter. 

    Does not add X-Forwarded-For or other standard headers.
    
    :param environ: The WSGi environ parameter passed to the application endpoint.
    :param start_response: The WSGi start_response passed to the application endpoint.
    :param gateway_url: A parsed gateway URL a returned by urllib.parse.urlparse()
    :param block_size: The block size used to transfer data from the gateway to the client.
    :return: The method acts as a generator function for the proxied HTTP response. 
    """
    
    if gateway_url.scheme == 'http':
        ConnClass = httplib.HTTPConnection
    elif gateway_url.scheme == 'https':
        ConnClass = httplib.HTTPSConnection
    else:
        raise ValueError(
            "Unknown scheme: %r" % gateway_url.scheme)
        
    conn = ConnClass(gateway_url.hostname,gateway_url.port)
    try:
        headers = {}
        for key, value in environ.items():
            if key.startswith('HTTP_'):
                key = key[5:].replace('_', '-').title()
                headers[key] = value
                
        path = (url_quote(environ.get('SCRIPT_NAME', ''))
                + url_quote(environ.get('PATH_INFO', '')))
        
        if environ.get('QUERY_STRING'):
            path += '?' + environ['QUERY_STRING']
        try:
            content_length = int(environ.get('CONTENT_LENGTH', '0'))
        except ValueError:
            content_length = 0
        if content_length:
            body = environ['wsgi.input'].read(content_length)
        else:
            body = ''
        headers['Content-Length'] = content_length
        if environ.get('CONTENT_TYPE'):
            headers['Content-Type'] = environ['CONTENT_TYPE']
        if not path.startswith("/"):
            path = "/" + path
        try:
            conn.request(environ['REQUEST_METHOD'],
                         path, body, headers)
        except socket.error as exc:
            if exc.args[0] == -2:
                # Name or service not known
                return send_error(environ,start_response,'502 Bad Gateway',"Hostname [%s] not found."%gateway_url.hostname)
            else:
                return send_error(environ,start_response,'503 Service Unavailable',"Upstream server [%s] failed to respond."%gateway_url.hostname)
        
        res = conn.getresponse()
        headers_in = res.getheaders()
        headers_out = [x for x in headers_in if not x[0].lower() in FILTERED_HEADERS]
        status = '%s %s' % (res.status, res.reason)
        
        start_response(status, headers_out)
    
        data = res.read(block_size)
    
        while len(data) > 0:
            yield data
            data = res.read(block_size)
    
    finally:
        conn.close()
