'''
..
  $Id$
 
  HTTP python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import types
import mimetypes
import io

NOCACHE_HEADERS = [("Cache-Control","no-cache"),
                   ("Pragma","no-cache"),
                   ("Expires","0")]

DEFAULT_BLOCK_SIZE=8192

def send_error(environ:map, start_response:types.FunctionType, status:str, msg:str):
    '''
    Send a simple error message reponse to the client.
    
    :param environ: The WSGi environ parameter passed to the application endpoint.
    :param start_response: The WSGi start_response passed to the application endpoint.
    :param status: SThe status string to pass to start_response alongside with the content-type.
    :param msg: The plain-text error message to be sent to the client.
    :return: The utf-8 encode error message as a list of bytes.
    '''
    start_response(status,[('Content-Type','text/plain;charset=utf-8')])
    return [msg.encode('utf-8')]
    
def send_filelike(environ:map, start_response:types.FunctionType, fh, headers:list,block_size=DEFAULT_BLOCK_SIZE):
    '''
    Send a file-like object with response code 200 and configurable headers as response
    to a WSGi request.
    
    :param environ: The WSGi environ parameter passed to the application endpoint.
    :param start_response: The WSGi start_response passed to the application endpoint.
    :param fh: A file-like handle to send to the client.
    :param headers: A list of header name/value pair to send as response headers.
    :param block_size: The block size used to transfer data from the gateway to the client.
    :return: The method returns an iterator to file-like object.
    '''
    
    try:
        start_response("200 OK",headers)
        
        if 'wsgi.file_wrapper' in environ:
            return environ['wsgi.file_wrapper'](fh, block_size)
        else:
            return iter(lambda: fh.read(block_size),b'')
    except:
        # if anything goes wrong above, close the handle, for  sure. 
        fh.close()

def send_file(environ:map, start_response:types.FunctionType, filename: str, headers:list,block_size=DEFAULT_BLOCK_SIZE):
    '''
    Send a file with response code 200 and configurable headers as response
    to a WSGi request.
    
    The ``Content-Type`` and ``Content-Encoding`` headers will be added to the passed in headers
    based on the results of mimetypes.guess_type(filename)
    
    :param environ: The WSGi environ parameter passed to the application endpoint.
    :param start_response: The WSGi start_response passed to the application endpoint.
    :param filename: The path of the file to send as the WSGi response.
    :param headers: A list of header name/value pair to send as response headers.
    :param block_size: The block size used to transfer data from the gateway to the client.
    :return: The method acts as a generator function for the content of the file-like object.
    '''
            
    mime_type,encoding = mimetypes.guess_type(filename)

    headers_final = list(headers) 
    
    if mime_type is not None:
        headers_final.append(("Content-Type",mime_type))
    
    if encoding is not None:
        headers_final.append(("Content-Encoding",encoding))

    try:    
        fh = io.open(filename,"rb")
        return send_filelike(environ,start_response,fh,headers_final,block_size)
    
    except IOError:
        return send_error(environ,start_response,'404 Not Found',"File [%s] not found."%filename)
