'''
..
  $Id$
  
  HTTP python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License. 
'''

import types

from clazzes.http.response import DEFAULT_BLOCK_SIZE, send_file, NOCACHE_HEADERS,\
    send_error
from urllib.parse import ParseResult
from clazzes.http.proxy import proxy_exact_request
import os
import stat

import logging

log = logging.getLogger(__name__)

class ProxyHandler:
    '''
    A handler, which delegates to a reverse proxy.
    :ivar gateway_url: The parsed upstream gateway URL. 
    '''
    def __init__(self, gateway_url: ParseResult):
        self.gateway_url = gateway_url

    def __str__(self):
        return "ProxyHandler[%s]" % self.gateway_url

    def isValid(self):
        '''
        :return: True, if the gateway URL is well-formed and contains a known HTTP scheme and a hostname.
        '''
        return (self.gateway_url and 
                (self.gateway_url.scheme == "http" or self.gateway_url.scheme == "https") and
                self.gateway_url.hostname is not None)             

    def handle_request(self,environ:map, start_response:types.FunctionType, sub:str, block_size=DEFAULT_BLOCK_SIZE):
        '''
        Process a request and delegate to the gateway server.
        
        :param environ: The WSGi environ parameter passed to the application endpoint.
        :param start_response: The WSGi start_response passed to the application endpoint.
        :param sub: The trailing pathinfo after the registered prefix of this handler.
        :param block_size: The block size to process data
        :return: The iterable data to be written to the WSGi response.
        '''
        return proxy_exact_request(environ,start_response,self.gateway_url,block_size)

class DirHandler:
    '''
    A handler, which sends a file from a given directory to WSGi response.
    :ivar dirpath: The path of the directory to serve the content from
    :ivar nocache: Whether to send the infamous nocache headers. 
    '''
    def __init__(self, dirpath: ParseResult, nocache = True):
        self.dirpath = dirpath
        self.nocache = nocache

    def __str__(self):
        return "DirHandler[%s]" % self.dirpath

    def isValid(self):
        '''
        :return: True, if the given directory path exists and is a directory. 
        '''
        try:
            st = os.stat(self.dirpath)
            return stat.S_ISDIR(st.st_mode)
        except FileNotFoundError:
            return False

    def handle_request(self,environ:map, start_response:types.FunctionType, sub:str, block_size=DEFAULT_BLOCK_SIZE):
        '''
        Process a request and server a file from the given path.
        
        :param environ: The WSGi environ parameter passed to the application endpoint.
        :param start_response: The WSGi start_response passed to the application endpoint.
        :param sub: The trailing pathinfo after the registered prefix of this handler.
        :param block_size: The block size to process data
        :return: The iterable data of the file to be written to the WSGi response.
        '''
        filename = os.path.join(self.dirpath,sub)
        headers = NOCACHE_HEADERS if self.nocache else []
        return send_file(environ,start_response,filename,headers,block_size)

def lookup_prefix(prefixes: map, pathinfo: str):
    '''
    Lookup a path handler in a prefix map by splitting the path by
    the included slashes.
    
    :param prefixes: A map of prefixes as keys with associated handlers as values.
                     The prefixes should end with a slash, if they map a whole subtree,
                     otherwise, they refer to an exact pathinfo. A fallback
                     handler may be provided as value of the empty path string.
    :param pathinfo: The value of the WSGi ``PATH_INFO`` variable.
    :return: A pair containing the matcher handler and the trailing part of
            the pathinfo after the path of the handler in the prefix map. 
    '''
    pos = len(pathinfo)
    
    # try exact match
    res = prefixes.get(pathinfo)
    sub = None
    
    try:
        
        while res is None and pos > 0:
        
            # descend down the tree.
            pos = pathinfo.rindex('/',0,pos)
            res = prefixes.get(pathinfo[:pos+1])
            sub = pathinfo[pos+1:]
        
    except ValueError:
        pass

    if res is None:
        res = prefixes.get('')
        sub = pathinfo
    
    return (res,sub)

def validate_prefixes(prefixes, fallback_handler=None):
    '''
    Validate configured prefix handlers and skip prefixes with invalid handlers,
    which will be replaced by ``fallback_handler``, if not None
    
    :param prefixes: A map of prefixes to URL handlers.
    :param fallback_handler: An optional handler, which will replace invalid handlers.
    :return: A map prefixes with valid URL handlers only.
    '''
    
    ret = {}
    
    for pfx,handler in prefixes.items():
        if not handler.isValid():
            log.warn("Handler [%s] for [%s] is invalid, skipping it..."%(handler,pfx))
            if fallback_handler is not None:
                ret[pfx] = fallback_handler
        else:
            ret[pfx] = handler
    return ret

def serve_site(environ:map, start_response:types.FunctionType, prefixes: map, aliases = None, block_size=DEFAULT_BLOCK_SIZE):
    '''
    WSGi handler for a simple site-setup represented by a prefix map.
    
    :param environ: The WSGi environ parameter passed to the application endpoint.
    :param start_response: The WSGi start_response passed to the application endpoint.
    :param prefixes: A map of prefixes to URL handlers.
    :param aliases: An optional list of aliases for substituting alternate PATHs for given exact pathinfos
    :param block_size: The block size used to transfer data from the gateway to the client.
    :return: The method returns an iterator to file-like object.
    '''
    
    pi = environ["PATH_INFO"]
    
    if aliases is not None:
        alias = aliases.get(pi)
        if alias is not None:
            pi = alias
            environ["PATH_INFO"] = alias
    
    handler,sub = lookup_prefix(prefixes,pi)
    
    if handler is None:
        return send_error(environ,start_response,"404 Not Found","No handler path info [%s]"%pi)
    else:
        return handler.handle_request(environ,start_response,sub,block_size = block_size)
