'''
..
  $Id: site.py 90 2018-07-10 09:05:40Z pytools $
  
  HTTP python tools of the clazzes.org project
  http://www.clazzes.org
 
  Created: 20.09.2016
 
  Licensed under the Apache License, Version 2.0 (the "License");
  you may not use this file except in compliance with the License.
  You may obtain a copy of the License at
 
      http://www.apache.org/licenses/LICENSE-2.0
 
  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.
'''

import requests
import typing
import threading
import collections
import logging

from xml.dom.minidom import parseString, Document
from inspect import ismethod

log = logging.getLogger(__name__)

def getFieldValue(xmldoc : Document, fieldname : str):
    '''
    Extracts the value of a named field from an XML document containing a login form.
    
    :param xmldoc: The document to be searched
    :param fieldname: 
    :return: The page token as a string, or None if not found
    '''
    inputs = xmldoc.getElementsByTagName('input')
    for inp in inputs:
        if inp.getAttribute('name') == fieldname:
            return inp.getAttribute('value')
    return None
    

def getPageToken(xmldoc : Document):
    '''
    Extracts the page token from an XML document containing a login form.
    
    :param xmldoc: The document to be searched
    :return: The page token as a string, or None if not found
    '''
    return getFieldValue(xmldoc,'pageToken')

def getPrincipal(xmldoc: Document):
    '''
    Extracts the principal name from an XML document containing a login form.
    
    :param xmldoc: The document to be searched
    :return: The principal name as a string, or None if not found
    '''
    return getFieldValue(xmldoc,'principal')

class NoCookieException(Exception):
    pass

class NoPrincipalException(Exception):
    pass

class NoStatusException(Exception):
    pass

class NoPageTokenException(Exception):
    pass

STATUS_LOGIN_OK = 200
STATUS_UNAUTHENTICATED = 401
STATUS_FORBIDDEN = 403
STATUS_NOT_ACCEPTABLE = 406
STATUS_EXPECT_TOKEN_OTP = 491
STATUS_EXPECT_EPHEMERAL_OTP = 492

class LoginInfo:
    '''
       The information gained from a successful login attempt on a login service.
       
       :ivar principal: The name of the user possibly prefixed with a domain and a slash.
       :ivar cookies: A requests CookieJar instance as contained in the server's response to the log in.
       :ivar status: The login status.
       :ivar may_send_sms: Whether an SMS OTP my be sent to the user.
    '''
    
    __slots__ = ( 'principal','cookies','page_token','status','may_send_sms')
    
    def __init__(self,principal : str,cookies,page_token : str = None, status: int = None, may_send_sms:bool = None):
        self.principal = principal
        self.cookies = cookies
        self.page_token = page_token
        self.status = status
        self.may_send_sms = may_send_sms
        
    def __str__(self):
        return "LoginInfo[%s;%s;%s]"%(self.principal,self.cookies,self.status)
    

class LoginService:
    '''
    Service that, given a base url and a mechanism, will provide a means to log in at the URL assembled from
    the base url and the mechanism. The way the URL is assembled relies on the schema 
    (baseurl + '/http-login/' + mechanism + '/login'), which is hard-coded.
    
    :ivar baseurl: top-level URL of the host containing the Login-Service
    :ivar mechanism: The mechanism used for logging in. If this string start with a slash,
                     it is used as a relative URL for the login dialog.
    '''
    
    def getLoginUrl(self):
        
        if self.mechanism[:1] == '/':
            return self.baseurl+self.mechanism
        else:
            return self.baseurl+'/http-login/'+self.mechanism+'/login'
    
    def __init__(self,baseurl,mechanism):
        
        if baseurl[-1:] == '/':
            self.baseurl = baseurl[:-1]
        else:
            self.baseurl = baseurl
            
        self.mechanism = mechanism
    
    def _parseLoginForm(self,url,res, old_cookies = None):
        """
        Parses the XML document returned by the login Service after a login attempt and searches it for login cookies, 
        principal and the custom response code which may indicate an error.

        @param url: The URL the login service is located at.
        @param res: The class:`requests.request` object retrieved from the login service to the request attempting to log in.
        @param old_cookies: In case no new cookies are returned in the @res parameter if the old ones remain valid, these cookies
        are used.
        @return: The :class:`loginInfo` object which contains URL, cookies, principal and status returned by the login service.
        @raises NoPrincipalException: If no principal was found in @res. 
        @raises NoCookieException: If no cookie was found in @res. 
        @raises NoStatusException: If no status code was found in @res. 
        """
        
        xmldoc = parseString(res.content)
        token = getPageToken(xmldoc)

        try:
            loginCookies = res.cookies
        except KeyError:
            raise NoCookieException('Did not receive cookie from login service [%s]'%url)
        
        if not loginCookies and old_cookies is not None:
            loginCookies = old_cookies
        
        try:
            principal = getPrincipal(xmldoc)
        except KeyError:
            raise NoPrincipalException('Did not receive a principal from login service [%s]'%url)
        
        try:
            status = int(getFieldValue(xmldoc,'status'))
        except (KeyError,ValueError):
            raise NoStatusException('Did not receive a status from login service [%s]'%url)
        
        if status == STATUS_EXPECT_TOKEN_OTP:
        
            gso = getFieldValue(xmldoc,'generateSmsOtp')
        
            may_send_sms = gso is not None
        else:
            may_send_sms = None
        
        return LoginInfo(principal,loginCookies,token,status,may_send_sms)
    
    def performLogin(self,username,password):
        '''
        Attempts to log in with given username and password, at the url specified by the class attributes. 
        
        :return: A LoginInfo class instance with the appropriate principal and cookie attributes set
        :raise NoCookieException: if no cookie could be retrieved after attempting to log in 
        :raise NoPrincipalException: If no principal could be retrieved from the response body after attempting to log in
        '''
        res = requests.get(self.getLoginUrl())
        xmldoc = parseString(res.content)
        token = getPageToken(xmldoc)
        url = self.getLoginUrl()
        res = requests.post(url, 
                              data={'pageToken':token, 'user':username, 'password':password} )
        return self._parseLoginForm(url,res)
    
    def sendTokenOtp(self, login_info: LoginInfo, otp: str):
        """
        Send a token OTP to the login service URL saved in this class.
        
        @param loginInfo: The :class:`loginInfo` object containing the page token and cookies.
        @param otp: The One Time Password needed for two-factor authentication.
        @return: The response to the HTTP request that includes the token. 
        """
        url = self.getLoginUrl()
        res = requests.post(url, cookies = login_info.cookies, 
                              data={'pageToken':login_info.page_token, 'tokenOtp':otp} )
        return self._parseLoginForm(url,res,login_info.cookies)
    
    def generateSmsOtp(self, login_info: LoginInfo):
        """
        Request, from the login service, a SMS One Time Password to be sent in order to complete two-factor authentication. 
        @param loginInfo: The :class:`loginInfo` containing the cookies and page token.
        @return: The response to the HTTP request requesting the SMS token.
        """
        url = self.getLoginUrl()
        res = requests.post(url, cookies = login_info.cookies, 
                              data={'pageToken':login_info.page_token, 'generateSmsOtp':'true'} )
        return self._parseLoginForm(url,res,login_info.cookies)

    def sendEphemeralOtp(self, login_info: LoginInfo, otp: str):
        
        url = self.getLoginUrl()
        res = requests.post(url, cookies = login_info.cookies, 
                              data={'pageToken':login_info.page_token, 'ephemeralOtp':otp} )
        return self._parseLoginForm(url,res,login_info.cookies)

    def performLogout(self, login_info: LoginInfo):
        """
        Attempt to log out from the set login service. Cancel if either cookies, page token or principal are not set for it.
        
        @param loginInfo: The :class:`loginInfo` object holding all information about the login session to be logged out from.
        """
        url = self.getLoginUrl()
        
        if not hasattr(login_info, 'cookies') or login_info.cookies.get('LOGIN_SESSION_ID') == None:
            raise NoCookieException('LoginInfo object passed does not contain a login session cookie for URL [%s]'%url)
        if not hasattr(login_info, 'page_token') or login_info.page_token == '' or login_info.page_token == None:
            raise NoPageTokenException('LoginInfo object passed does not contain a page token for URL [%s]'%url)
        if not hasattr(login_info, 'principal') or login_info.principal == '' or login_info.principal == None:
            raise NoPrincipalException('LoginInfo object passed does not contain a principal for URL [%s]'%url)
        
        res = requests.post(url, cookies = login_info.cookies, 
                              data = {'pageToken':login_info.page_token, 'logout': 'true'})
        
        if log.getEffectiveLevel() <= logging.DEBUG:
            log.debug("logout from URL [%s] returned [%s]"%(url,res.content))

class ParameterInfo:
    '''
    Holds information about one parameter of a SMD method, extracted from a SMD .json message. Part of the ~xmlrpc.MethodInfo~ class.
    
    :ivar type: The required type for the parameter
    :ivar optional: Whether this parameter needs to be be sent.
    '''
    __slots__ = ("type","optional","enum")
    
    def __str__(self):
        if (self.optional == True):
            return '[%s]'%self.type
        else:
            return '%s'%self.type
    
    def __init__(self, type:str = '', optional:bool = False,enum:collections.abc.Sequence = None):  # @ReservedAssignment
        self.type = type
        self.optional = optional
        self.enum=enum
    
    def getDefault(self):
        '''
           :return: A default parameters to pass to the service for this positional parameter.
                    For optional parameters None will be returned, for mandatory parameters
                    an empty strings, zero or the first valid enum value will be returned.
        '''
        if self.optional:
            return None
        elif self.type == "number":
            return 0
        elif self.type == "boolean":
            return False
        elif self.enum:
            return self.enum[0]
        else:
            return ""

class MethodInfo:
    '''
    Holds information about a SMD Method, extracted from a SMD .json message. Part of the ~xmlrpc.SMDInfo~ class.
    
    :ivar parameters: The methods' parameters stored as ~xmlrpc.ParameterInfo~ intances.
    '''
    __slots__ = ("parameters")
    
    def __str__ (self):
        retstr = '('
        for p in self.parameters:
            retstr += p.__str__()
            if self.parameters[-1] == p:
                continue
            retstr += ', '
        retstr += ')'
        return retstr
    
    def __init__(self, parameters:typing.Sequence):
        self.parameters = [ ParameterInfo(**p) for p in parameters]

    def getDefaultParameters(self):
        '''
           :return: A tuple of default parameters to pass to the service in order to
                    perform a syntactically correct RPC call. Optional parameters will
                    be represented with a None value, mandatory parameters will be filled
                    with empty strings, zeros or the first valid enum value.
        '''
        return tuple(p.getDefault() for p in self.parameters)
    
class SMDInfo:    
    '''
    ServiceMethodDescription Info.
    Contains information about available methods of a service for RPC.  
    Usually created by SMDService after parsing the contents of the smd json
    
    :ivar smd_ver: The SMD version of the recieved .json message(usually 2.0)
    :ivar smdid: The unique ID of the service. Defaults to the complete (relative) URL to it (including ?smd ending)
    :ivar envelope: The method of transportation used for passing the method parameters to the service. 
                    Default is JSON_RPC_2.0, which is sent in the message body of a POST request as a json.
    :ivar target: The path to the service that created the .json message (without ?smd ending)
    :ivar transport: The method that should be used to execute service calls. 
    :ivar services: A multi-level dictionary containing the names of the methods as key and the available
                    methods for each of them as value. Methods are also represented as a dictionary containing
                    the needed parameters, their type and optionality.
    
    
    '''
    __slots__ = ("smd_ver","smdid","envelope","target","transport","services")
    
    def __init__(self, SMDVersion : str, id : str, envelope: str, target : str, transport : str, services : dict):  # @ReservedAssignment
        self.smd_ver = SMDVersion
        self.smdid = id
        self.envelope = envelope
        self.target = target
        self.transport = transport
        self.services = { name : MethodInfo(**m) for name,m in services.items() }

    def __str__(self):
        return "SMDInfo[%s - %s: %s]"%(self.target, self.smdid, self.services.__str__())
    
    def getBeanName(self):
        nm = self.smdid
        for c in '?/-':
            nm = nm.replace(c,'_')
        return nm
    
JSON_RPC_PARSE_ERROR = -32700
JSON_RPC_INVALID_REQUEST = -32600
JSON_RPC_METHOD_NOT_FOUND = -32601
JSON_RPC_INVALID_PARAMS = -32602
JSON_RPC_INTERNAL_ERROR = -32603
JSON_RPC_SERVER_ERROR_0 = -32000
JSON_RPC_LOGIN_REQUIRED = -32099
         
class XmlRpcProtocolException(Exception):
    pass

class XmlRpcTargetError(Exception):
    
    __slots__ = ( 'code','data','baseurl' )
    
    def __init__(self,code : int,message : str,data = None, baseurl = None):
        
        super().__init__(message)
        self.code = code
        self.data = data
        self.baseurl = baseurl

class RPCService:
    '''
    Service for remotely accessing a service on the server. 
    Allows for retrieving a smd .json message from a service, and parsing it into a ~xmlrpc.SMDInfo~ instance
    containing information about their available methods.
    Capable of creating a bean from the SMDInfo providing the remote service's methods, which are accessed
    via RPC.
    
    :ivar baseurl: The top-level url of the host operating the service
    :ivar service: The name of the service
    :ivar lastid: The id of the most recent request sent to the server. Starts at 0 and increments
    '''
    
    __slots__ = ( "baseurl", "service", "lastid")
    
    def getServiceUrl(self):
        return self.baseurl+self.service
    
    def __init__(self, baseurl : str, service : str): 
        if baseurl[-1:] == '/':
            self.baseurl = baseurl[:-1]
        else:
            self.baseurl = baseurl
        self.service = service
        self.lastid = 0
    
    def __str__(self):
        return self.service
    
    def __repr__(self):
        return "RpcService[baseurl=%s,service=%s]"%(self.baseurl,self.service)

    def retrieveSMDInfo(self):
        '''
        Accesses the service and retrieves its SMD data.
        
        :return: A ~xmlrpc.SMDInfo~ instance extracted from the received .json message.
        :raise ValueError: In case the result is not decipherable by requests .json() method.
        '''
        url = self.getServiceUrl() + '?smd'
        result = requests.get(url)
        smdValues = result.json()
        return SMDInfo(**smdValues)

    def callMethod(self, name : str, *args, login_info:LoginInfo=None, login_info_factory=None):
        
        if login_info_factory is not None:
            login_info = login_info_factory()
        
        self.lastid += 1
        # req = { 'id' : self.lastid, 'jsonrpc': '2.0', 'method': name, 'params': args}
        # work around a parsing problem in JsonRpcServlet as of version 1.1.1, which
        # unfortunately depends on the order of parsed items in the request.
        req = collections.OrderedDict((('id',self.lastid),('jsonrpc','2.0'),('method',name),('params',args)))
        response = requests.request('POST',self.getServiceUrl(),json=req, cookies=None if login_info is None else login_info.cookies)
        response.raise_for_status()
        resp = response.json()
        
        if resp.get('jsonrpc') != '2.0':
            raise XmlRpcProtocolException("Received a non JSON-RPC 2.0 response.")

        if resp['id'] != self.lastid:
            raise XmlRpcProtocolException("Request ID [%d] does not match response ID [%d]"%(self.lastid,resp['id']))
        
        err = resp.get('error')
        
        if err is not None:
            raise XmlRpcTargetError(err.get('code'),err.get('message'),err.get('data'),self.baseurl)
        
        return resp.get('result')
            

bean_type_lock = threading.Lock()
bean_types = {}

def defineMethod(ni):
    '''
    Returns a function object containing the call to the corresponding service method on the server.
    'self' usually references the dynamically created class (see ~xmlrpc.getBeanClass), which receives the returned
    function of this method and maps it into its name space.
    
    :param ni: A pair of a name and the ~xmlrpc.MethodInfo of the method as noted in the SMD,
               of which the function proxy should be returned.
    '''
    
    def callRpcMethod(self,*args):
        return self._service.callMethod(ni[0],*args,login_info = self._login_info, login_info_factory = self._login_info_factory)

    return callRpcMethod

def methodName(func):
    '''
    :param func: A python method, possibly returned by ~xmlrpc.defineMethod
    :return: The name of the given method, possibly the reolved name of the rmote method to be called.
    '''
    if (ismethod(func) and func.__name__ == "callRpcMethod" and
        len(func.__code__.co_freevars) > 0 and func.__code__.co_freevars[0] == "ni" and
        func.__closure__):
        return func.__closure__[0].cell_contents[0]
    
    return func.__name__

def methodInfo(func):
    '''
    :param func: A python method, possibly returned by ~xmlrpc.defineMethod
    :return: The ~xmlrpc.MethodInfo of the given method, or None, if the method
             has not been constructed by ~xmlrpc.defineMethod .
    '''
    if (ismethod(func) and func.__name__ == "callRpcMethod" and
        len(func.__code__.co_freevars) > 0 and func.__code__.co_freevars[0] == "ni" and
        func.__closure__):
        return func.__closure__[0].cell_contents[1]
    
    return None

def getBeanClass(smd_info : SMDInfo):
    '''
    Create, dynamically, a RPC class that contains all the methods of the corresponding remote server.
    Depends on the global ~xmlrpc.beanTypes~ dictionary to save the names of already created beans.
    Will only create a class from the same service URL once.
    Uses type() method for class creation.
    Is called in ~xmlrpc.makeBean~ function which also assigns it the loginInfo object needed for 
    executing the RPC, making it fully functinal.
    
    :param smd_info: A ~xmlrpc.SMDInfo~ instance containing information about the remote service.
    :return: A class object containing the functions that were found in the SMD, which forward the call
             to those remote service functions when called. 
    '''

    typeName = smd_info.getBeanName()

    with bean_type_lock:

        #Will not create class if already registered in global BeanTypes collection
        BeanClass = bean_types.get(typeName)

        if BeanClass is None:

            classDict = {}

            #manually set name space for to-be-created bean. defineMethod() returns a function object which is mapped
            #to the function's name.
            for ni in smd_info.services.items():
                classDict[ni[0]] = defineMethod(ni)

            #Define constructor for new class so that _service and _login_info are always set
            def stubCtor(self,service,login_info, login_info_factory=None):
                self._service = service
                self._login_info = login_info
                self._login_info_factory = login_info_factory
            
            classDict["__init__"] = stubCtor

            def stubStr(self):
                return smd_info.target

            classDict["__str__"] = stubStr

            #Creates the class object. Second parameter is the tuple of base classes, saved in __bases__  variable of the
            #newly created class.
            #Third parameter the name space to be set for the class, which contains the methods.
            BeanClass = type(typeName,(object,),classDict)

            bean_types[typeName] = BeanClass

    return BeanClass

def makeBean(service : RPCService, smd_info : SMDInfo, login_info:LoginInfo=None, login_info_factory=None):
    '''
    Creates a class object from the given SMD, which acts as a representant of the remote service.
    Assigns it a loginInfo object needed to make RPC calls.
    See ~xmlrpc.getBeanClass~ for details on bean creation.
    
    :param RPCService: The RPCService instance of the service which should be represented.
    :param smd_info: Contains the SMD of the methods which should be created.
    :param login_info: An ~xmlrpc.LoginInfo~ object needed to make RPC calls to the remote service.
    :param login_info_factory: A factory method, which returns an ~xmlrpc.LoginInfo~ object, which
                       is an alternative to specifying a login_info directly. If both login_info
                       and login_info_factory are given, login_info_factory takes precedence.
    '''

    BeanClass = getBeanClass(smd_info)

    return BeanClass(service,login_info,login_info_factory)
